import seaborn as sns


palette = sns.color_palette('colorblind')


class Colours:
    shade_color = palette[0]
    line_color = palette[0]
    line2_color = palette[2]
    shade2_color = palette[2]
    scatter_color = palette[3]
    bar1_color = palette[3]
    bar2_color = palette[2]
    error_color = palette[1]
